/*
    Project: TCTEMP
    Version: V1.18
    File   : na_image.c

    Copyright (c) 2006-2009 Author of TCTEMP. All rights reserved.
    Public Key Fingerprint: 75EB 6BC2 01B7 F6E7 4BD7 CC58 4A5F C393 19EE 6E69

    Covered by TrueCrypt License 2.3 the full text of which is contained
    in the file License.txt included in TCTEMP binary and source code
    distribution archives.

    The copyright notice of this source code file and the public key
    fingerprint must be retained in full.
*/

#include "na.h"
#include "na_image.h"
#include <ntdddisk.h>

#define NaReadSector(hFile, pSector, pliOffset, uSectorSize) \
    NaReadFile(hFile, pSector, uSectorSize, NULL, pliOffset)

#define NaWriteSector(hFile, pSector, pliOffset, uSectorSize) \
    NaWriteFile(hFile, pSector, uSectorSize, NULL, pliOffset)

static BOOL NAAPI _NaRestoreFileSystem(PIMAGE_CONTEXT pCtx)
{
    PIMAGE_HEADER pih;
    PIMAGE_ENTRY  pie;
    PIMAGE_ENTRY  pieBegin;
    PIMAGE_ENTRY  pieEnd;
    LARGE_INTEGER li;
    LARGE_INTEGER liDrivePos;
    ULONG         uSectorSize;
    DISK_GEOMETRY dg;
    ULONG         ul;

    if (!NaReadSector(pCtx->hHeader, pCtx->rgbHeaderSector, NULL, 512)) {
        NaDbgPrintf(L"Cannot read image file header!");
        return FALSE;
    }
    pih = (PIMAGE_HEADER)pCtx->rgbHeaderSector;
    if (pih->dwId != IMAGE_HEADER_ID) {
        NaDbgPrintf(L"Invalid image file (wrong header id)!");
        return FALSE;
    }

    if (pih->wSectorSize != 512 &&
      pih->wSectorSize != 1024 &&
      pih->wSectorSize != 2048 &&
      pih->wSectorSize != 4096 &&
      pih->wSectorSize != 8192) {
        NaDbgPrintf(L"Unsupported image file (sector size = %d)!",
          (int)pih->wSectorSize);
        return FALSE;
    }
    if (pih->ullEntries == 0) {
        NaDbgPrintf(L"Unsupported image file (no entries)!");
        return FALSE;
    }
    uSectorSize = pih->wSectorSize;

    if (!NaDeviceIoControl(pCtx->hDrive, IOCTL_DISK_GET_DRIVE_GEOMETRY,
      NULL, 0, &dg, sizeof(dg), &ul) || ul != sizeof(dg)) {
        NaDbgPrintf(L"Cannot get drive geometry!");
        return FALSE;
    }
    if (dg.BytesPerSector != pih->wSectorSize) {
        NaDbgPrintf(L"Disk sector size is different to disk image!");
        return FALSE;
    }
    li.QuadPart = dg.Cylinders.QuadPart * dg.SectorsPerTrack *
      dg.TracksPerCylinder;
    if (li.QuadPart != pih->ullTotalSectors) {
        NaDbgPrintf(L"Disk size is different to disk image!");
        return FALSE;
    }

    li.QuadPart = uSectorSize;
    if (!NaSeek(pCtx->hHeader, &li)) {
        NaDbgPrintf(L"Cannot move file pointer to first entry!");
        return FALSE;
    }

    if (!NaSeek(pCtx->hData, (PLARGE_INTEGER)&pih->ullDataByteOffset)) {
        NaDbgPrintf(L"Cannot move file pointer to first data block!");
        return FALSE;
    }

    liDrivePos.QuadPart = uSectorSize;
    if (!NaReadSector(pCtx->hHeader, pCtx->rgbHeaderSector, &liDrivePos,
      uSectorSize)) {
        NaDbgPrintf(L"Cannot read first entry sector!");
        return FALSE;
    }

    pie = pieBegin = (PIMAGE_ENTRY)pCtx->rgbHeaderSector;
    pieEnd = pieBegin + IMAGE_ENTRIES_PER_SECTOR(uSectorSize);

    while (pie->wType) {
        if (HIBYTE(pie->wType) == HIBYTE(IT_FILL)) {
            memset(pCtx->rgbDataSector, LOBYTE(pie->wType), uSectorSize);
            if (!NaWriteSector(pCtx->hDrive, pCtx->rgbDataSector,
              (PLARGE_INTEGER)&pie->ullOffset, uSectorSize)) {
                return FALSE;
            }
            liDrivePos.QuadPart = pie->ullOffset + uSectorSize;
        } else {
            switch(pie->wType) {
            case IT_COPY:
                if (!NaReadSector(pCtx->hData, pCtx->rgbDataSector,
                  NULL, uSectorSize)) {
                    NaDbgPrintf(L"Cannot read image data sector!");
                    return FALSE;
                }
                if (!NaWriteSector(pCtx->hDrive, pCtx->rgbDataSector,
                  (PLARGE_INTEGER)&pie->ullOffset, uSectorSize)) {
                    return FALSE;
                }
                liDrivePos.QuadPart = pie->ullOffset + uSectorSize;
                break;
            case IT_MULTIPLE_COPY:
                while (pie->ullOffset) {
                    if (!NaReadSector(pCtx->hData, pCtx->rgbDataSector,
                      NULL, uSectorSize)) {
                        return FALSE;
                    }
                    if (!NaWriteSector(pCtx->hDrive, pCtx->rgbDataSector,
                      &liDrivePos, uSectorSize)) {
                        return FALSE;
                    }
                    liDrivePos.QuadPart += uSectorSize;
                    pie->ullOffset--;
                }
                break;
            case IT_REPEAT:
                while (pie->ullOffset) {
                    if (!NaWriteSector(pCtx->hDrive, pCtx->rgbDataSector,
                      &liDrivePos, uSectorSize)) {
                        return FALSE;
                    }
                    liDrivePos.QuadPart += uSectorSize;
                    pie->ullOffset--;
                }
                break;
            default:
                NaDbgPrintf(L"Invalid image file entry!");
                return FALSE;
            }
        }
        if (++pie == pieEnd) {
            if (!NaReadSector(pCtx->hHeader, pCtx->rgbHeaderSector,
              NULL, uSectorSize)) {
                return FALSE;
            }
            pie = pieBegin;
        }
    }

    return TRUE;
}

BOOL NAAPI NaRestoreFileSystem(PIMAGE_CONTEXT pCtx)
{
    BOOL fResult;

    fResult = FALSE;
    if (NaOpenFile(&pCtx->hHeader, pCtx->wszNtImageFileName, GENERIC_READ,
      FILE_SHARE_READ, FILE_SEQUENTIAL_ONLY|FILE_SYNCHRONOUS_IO_NONALERT)) {
        if (NaOpenFile(&pCtx->hData, pCtx->wszNtImageFileName, GENERIC_READ,
          FILE_SHARE_READ, FILE_SEQUENTIAL_ONLY|FILE_SYNCHRONOUS_IO_NONALERT)) {
            if (NaOpenFile(&pCtx->hDrive, pCtx->wszNtDriveName,
              GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE,
              FILE_SEQUENTIAL_ONLY|FILE_SYNCHRONOUS_IO_NONALERT)) {
                if (NaDeviceIoControl(pCtx->hDrive, FSCTL_LOCK_VOLUME,
                  NULL, 0, NULL, 0, NULL)) {
                    NaDeviceIoControl(pCtx->hDrive, FSCTL_ALLOW_EXTENDED_DASD_IO,
                      NULL, 0, NULL, 0, NULL);
                    fResult = _NaRestoreFileSystem(pCtx);
                    NaDeviceIoControl(pCtx->hDrive, FSCTL_DISMOUNT_VOLUME,
                      NULL, 0, NULL, 0, NULL);
                    NaDeviceIoControl(pCtx->hDrive, FSCTL_UNLOCK_VOLUME,
                      NULL, 0, NULL, 0, NULL);
                }
                NaCloseFile(pCtx->hDrive);
            }
            NaCloseFile(pCtx->hData);
        }
        NaCloseFile(pCtx->hHeader);
    }

    return fResult;
}
